# Scola version 1.1.0
# ScolaUniverse.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of ScolaWizard.
# ScolaWizard is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# ScolaWizard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with ScolaWizard.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Universe
# 3) Base types
# 4) Semi-lattices

# 1) Imported modules
# -------------------

import string
import sys
import math

from ScolaConstants import *

# 2) Universe
# -----------

# 2.1) Creator
# -----------

class ScolaUniverse:
  _me = None

  def __init__(self):
    self._noneType = None
    self._BooleanType = None
    self._integerType = None
    self._realType = None
    self._symbolType = None
    self._stringType = None
    self._anyType = None
    self._domains = dict()
    self._symbolicConstantsOfDomains = dict()

  @staticmethod
  def GetMe():
    if ScolaUniverse._me==None:
      ScolaUniverse._me = ScolaUniverse()
    return ScolaUniverse._me

  @staticmethod
  def Reset():
    me = ScolaUniverse.GetMe()
    me._noneType = None
    me._BooleanType = None
    me._integerType = None
    me._realType = None
    me._symbolType = None
    me._stringType = None
    me._anyType = None
    me._domains = dict()
    me._symbolicConstantsOfDomains = dict()

# 2.2) Base types
# ---------------

  @staticmethod
  def GetNoneType():
    me = ScolaUniverse.GetMe()
    if me._noneType==None:
      me._noneType = ScolaBaseType(ScolaBaseType.NONE, "none", me)
    return me._noneType

  @staticmethod
  def GetBooleanType():
    me = ScolaUniverse.GetMe()
    if me._BooleanType==None:
      me._BooleanType = ScolaBaseType(ScolaBaseType.BOOLEAN, "Boolean", me)
    return me._BooleanType

  @staticmethod
  def GetIntegerType():
    me = ScolaUniverse.GetMe()
    if me._integerType==None:
      me._integerType = ScolaBaseType(ScolaBaseType.INTEGER, "integer", me)
    return me._integerType

  @staticmethod
  def GetRealType():
    me = ScolaUniverse.GetMe()
    if me._realType==None:
      me._realType = ScolaBaseType(ScolaBaseType.REAL, "real", me)
    return me._realType

  @staticmethod
  def GetSymbolType():
    me = ScolaUniverse.GetMe()
    if me._symbolType==None:
      me._symbolType = ScolaBaseType(ScolaBaseType.SYMBOL, "symbol", me)
    return me._symbolType

  @staticmethod
  def GetStringType():
    me = ScolaUniverse.GetMe()
    if me._stringType==None:
      me._stringType = ScolaBaseType(ScolaBaseType.STRING, "string", me)
    return me._stringType

  @staticmethod
  def GetAnyType():
    me = ScolaUniverse.GetMe()
    if me._anyType==None:
      me._anyType = ScolaBaseType(ScolaBaseType.ANY, "any", me)
    return me._anyType

  @staticmethod
  def LookForDomain(identifier):
    me = ScolaUniverse.GetMe()
    return me._domains.get(identifier, None)

  @staticmethod
  def GetDomain(identifier):
    domain = ScolaUniverse.LookForDomain(identifier)
    if domain==None:
      me = ScolaUniverse.GetMe()
      domain = ScolaBaseType(ScolaBaseType.DOMAIN, identifier, me)
      me._domains[identifier] = domain
    return domain

  @staticmethod
  def LookForSymbolicConstantOfDomains(symbol):
    me = ScolaUniverse.GetMe()
    return me._symbolicConstantsOfDomains.get(symbol, None)

  @staticmethod
  def AddSymbolicConstantToDomain(symbol, domain):
    constant = ScolaUniverse.LookForSymbolicConstantOfDomains(symbol)
    if constant==None:
      me = ScolaUniverse.GetMe()
      constant = ScolaConstant.NewSymbol(symbol)
      me._symbolicConstantsOfDomains[symbol] = constant
      domain._members.append(constant)
    elif not constant in domain._members:
      domain._members.append(constant)
    return constant

# 3) Base types
# -------------

class ScolaBaseType:
  NONE = "NONE"
  BOOLEAN = "BOOLEAN"
  INTEGER = "INTEGER"
  REAL = "REAL"
  SYMBOL = "SYMBOL"
  STRING = "STRING"
  ANY = "ANY"
  DOMAIN = "DOMAIN"

  def __init__(self, type, identifier, universe):
    self._type = type
    self._identifier = identifier
    self._universe = universe
    self._members = []
    self._semiLattice = None

  @staticmethod
  def NewNone():
    return ScolaUniverse.GetNoneType()

  @staticmethod
  def NewBoolean():
    return ScolaUniverse.GetBooleanType()

  @staticmethod
  def NewInteger():
    return ScolaUniverse.GetIntegerType()

  @staticmethod
  def NewReal():
    return ScolaUniverse.GetRealType()

  @staticmethod
  def NewSymbol():
    return ScolaUniverse.GetSymbolType()

  @staticmethod
  def NewString():
    return ScolaUniverse.GetStringType()

  @staticmethod
  def NewAny():
    return ScolaUniverse.GetAnyType()

  @staticmethod
  def NewDomain(identifier):
    return ScolaUniverse.GetDomain(identifier)

  def GetType(self):
    return self._type
  def IsNone(self):
    return self._type==ScolaBaseType.NONE
  def IsBoolean(self):
    return self._type==ScolaBaseType.BOOLEAN
  def IsInteger(self):
    return self._type==ScolaBaseType.INTEGER
  def IsReal(self):
    return self._type==ScolaBaseType.REAL
  def IsSymbol(self):
    return self._type==ScolaBaseType.SYMBOL
  def IsString(self):
    return self._type==ScolaBaseType.STRING
  def IsAny(self):
    return self._type==ScolaBaseType.ANY
  def IsDomain(self):
    return self._type==ScolaBaseType.DOMAIN

  def GetIdentifier(self):
    return self._identifier

  def Contains(self, constant):
    if self.IsNone():
      return constant.IsNone()
    if self.IsBoolean():
      return constant.IsBoolean()
    if self.IsInteger():
      return constant.IsInteger()
    if self.IsReal():
      return constant.IsNumber()
    if self.IsSymbol():
      return constant.IsSymbol()
    if self.IsString():
      return constant.IsString()
    if self.IsAny():
      return True
    if self.IsDomain():
      return self.GetMember(constant)!=None
    return False # should not happen

  def AddSymbolicConstant(self, symbol):
    if not self.IsDomain():
      return
    return ScolaUniverse.AddSymbolicConstantToDomain(symbol, self)
  def AddMember(self, constant):
    if not self.IsDomain() or not constant.IsSymbol():
      return
    return ScolaUniverse.AddSymbolicConstantToDomain(constant.GetValueAsSymbol(), self)
  def GetMembers(self):
    return self._members
  def GetSymbolicConstant(self, symbol):
    if not self.IsDomain():
      return None
    symbolicConstant = ScolaUniverse.LookForSymbolicConstantOfDomains(symbol)
    if symbolicConstant==None:
      return None
    if symbolicConstant in self._members:
      return symbolicConstant
    return None
  def GetMember(self, constant):
    if not self.IsDomain() or not constant.IsSymbol():
      return None
    symbolicConstant = ScolaUniverse.LookForSymbolicConstantOfDomains(constant.GetValueAsSymbol())
    if symbolicConstant==None:
      return None
    if symbolicConstant in self._members:
      return symbolicConstant
    return None

  def GetSemiLattice(self):
    return self._semiLattice
  def AddOrderConstraint(self, constant1, constant2):
    if self.GetType()!=ScolaBaseType.DOMAIN:
      return False
    if self._semiLattice==None:
      self._semiLattice = ScolaSemiLattice(self)
    return self._semiLattice.AddOrderConstraint(constant1, constant2)
  def IsLessThan(self, constant1, constant2):
    if self.GetType()!=ScolaBaseType.DOMAIN or self._semiLattice==None:
      return False
    return self._semiLattice.IsLessThan(constant1, constant2)
  def GetLeastMembers(self):
    if self.GetType()!=ScolaBaseType.DOMAIN or self._semiLattice==None:
      return []
    return self._semiLattice.GetLeastMembers()

# 3) Semi-lattices
# ----------------

# 3.1) Order constraints
# ----------------------

class ScolaOrderConstraint():
  def __init__(self, member1, member2):
    self._member1 = member1
    self._member2 = member2

  def GetMember1(self):
    return self._member1
  def GetMember2(self):
    return self._member2

# 3.2) Semi-lattices
# ------------------

class ScolaSemiLattice():
  def __init__(self, enumeration):
    self._enumeration = enumeration
    self._constraints = []
    self_leastElement = None

  def IsLessThan(self, constant1, constant2):
    member1 = self._enumeration.GetMember(constant1)
    member2 = self._enumeration.GetMember(constant2)
    if member1==None or member2==None:
      return False
    return self._IsLessThan(member1, member2)

  def _IsLessThan(self, member1, member2):
    for constraint in self._constraints:
      if member1==constraint.GetMember1() and member2==constraint.GetMember2():
        return True
    return False

  def AddOrderConstraint(self, constant1, constant2):
    member1 = self._enumeration.GetMember(constant1)
    member2 = self._enumeration.GetMember(constant2)
    if member1==None or member2==None or member1==member2:
      return False
    if self._IsLessThan(member1, member2):
      return True
    if self._IsLessThan(member1, member2):
      return False
    self._constraints.append(ScolaOrderConstraint(member1, member2))
    self._CompleteOrderRelation(member1, member2)
    return True

  def _GetLowerMembers(self, member):
    lowerMembers = []
    for constraint in self._constraints:
      if member==constraint.GetMember2():
        lowerMembers.append(constraint.GetMember1())
    return lowerMembers

  def _GetGreaterMembers(self, member):
    greaterMembers = []
    for constraint in self._constraints:
      if member==constraint.GetMember1():
        greaterMembers.append(constraint.GetMember2())
    return greaterMembers

  def _CompleteOrderRelation(self, member1, member2):
    for member in self._GetLowerMembers(member1):
      if not self._IsLessThan(member, member2):
        self._constraints.append(ScolaOrderConstraint(member, member2))
    for member in self._GetGreaterMembers(member2):
      if not self._IsLessThan(member1, member):
        self._constraints.append(ScolaOrderConstraint(member1, member))

  def GetLeastMembers(self):
    leastMembers = self._enumeration.GetMembers().copy()
    for constraint in self._constraints:
      member = constraint.GetMember2()
      if member in leastMembers:
        leastMembers.remove(member)
    return leastMembers

  def GetOrderConstraints(self):
    return self._constraints

