# Scola version 1.1.0
# ScolaRewriter.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of ScolaWizard.
# ScolaWizard is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# ScolaWizard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with ScolaWizard.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Rewriter

# 1) Imported modules
# -------------------

import string
import sys
import re
from ScolaMetaModel import *
from ScolaModels import *
from ScolaPrinter import *
from ScolaExpressionEvaluator import *

# 2) Actions
# ----------

class ScolaRewriterAction:
  INSTANTIATE_ELEMENT = "INSTANTIATE_ELEMENT"

  TODO = "TODO"
  POSTPONED = "POSTPONED"
  COMPLETED = "COMPLETED"
  FAILED = "FAILED"

  def __init__(self, type, size):
    self._type = type
    self._status = ScolaRewriterAction.TODO
    self._size = size
    self._data = [None for _ in range(0, size)]

  def GetType(self):
    return self._type

  def GetStatus(self):
    return self._status
  def SetStatus(self, status):
    self._status = status

  def GetSize(self):
    return self._size
  def GetDatum(self, index):
    if index<0 or index>=self.GetSize():
      return None
    return self._data[index]
  def SetDatum(self, index, datum):
    if index<0 or index>=self.GetSize():
      return
    self._data[index] = datum

  @staticmethod
  def NewInstantiateElement(sourceElement, targetContainer):
    self = ScolaRewriterAction(ScolaRewriterAction.INSTANTIATE_ELEMENT, 2)
    self.SetDatum(0, sourceElement)
    self.SetDatum(1, targetContainer)
    return self

# 3) Rewriter
# -----------

# 3.1) Creator
# ------------

class ScolaRewriter:
  def __init__(self):
    self._sourceModel = None
    self._targetModel = None
    self._actionStack = []
    self._errors = 0
    self._noPostponing = False
    self._printer = ScolaPrinter()
    self._printer.SetOutputFile(sys.stderr)
    self._evaluator = ScolaExpressionEvaluator()

# 3.2) Error management
# ---------------------

  def GetErrorFile(self):
    return self._printer.GetOutputFile()
  def SetErrorFile(self, errorFile):
    self._printer.SetOutputFile(errorFile)
  def GetNumberOfErrors(self):
    return self._errors

  def RaiseError(self, message):
    self._errors += 1
    if self._errors>1:
      self._printer.Println()
    self._printer.Println("Error %d\n" % self._errors)
    self._printer.Println(message)
    self._printer.Flush()

# 3.3) Action stack
# -----------------

  def _IsActionStackEmpty(self):
    return len(self._actionStack)==0

  def _PushAction(self, action):
    self._actionStack.insert(0, action)

  def _EnqueueAction(self, action):
    self._actionStack.append(action)

  def _DequeueStack(self):
    return self._actionStack.pop(0)


# 3.4) Model expansion
# --------------------

  def CompileModel(self, sourceModel):
    self._sourceModel = sourceModel
    self._targetModel = ScolaModel()
    self._errors = 0
    self._InstantiateSourceModel()
    if self._errors!=0:
      return None
    path = ScolaPath()
    self._SetSystemPaths(self._targetModel, path)
    return self._targetModel

  def _InstantiateSourceModel(self):
    action = ScolaRewriterAction.NewInstantiateElement(self._sourceModel, self._targetModel)
    firstPostponedAction = None
    self._PushAction(action)
    self._noPostponing = False
    while not self._IsActionStackEmpty():
      action = self._DequeueStack()
      status = self._PerformAction(action)
      if status==ScolaRewriterAction.COMPLETED:
        firstPostponedAction = None
      elif status==ScolaRewriterAction.POSTPONED:
        self._EnqueueAction(action)
        if action==firstPostponedAction:
          break
        else:
          action.SetStatus(ScolaRewriterAction.POSTPONED)
          if firstPostponedAction==None:
            firstPostponedAction = action
      elif status==ScolaRewriterAction.FAILED:
        break
    self._noPostponing = True
    while not self._IsActionStackEmpty():
      action = self._DequeueStack()
      status = self._PerformAction(action)


  def _PerformAction(self, action):
    if action.GetType()==ScolaRewriterAction.INSTANTIATE_ELEMENT:
      return self._InstantiateElement(action)
    return ScolaRewriterAction.FAILED

  def _InstantiateElement(self, action):
    sourceElement = action.GetDatum(0)
    if sourceElement.GetType()==ScolaModel.MODEL:
      status = self._InstantiateModel(action)
    elif sourceElement.GetType()==ScolaModel.CLASS:
      status = ScolaRewriterAction.COMPLETED
    elif sourceElement.GetType()==ScolaModel.INSTANCE:
      status = self._InstantiateInstance(action)
    elif sourceElement.GetType()==ScolaModel.CLONES:
      status = self._InstantiateClonesDirective(action)
    elif sourceElement.GetType()==ScolaModel.INHERITS:
      status = self._InstantiateInheritsDirective(action)
    elif sourceElement.GetType()==ScolaModel.DOMAIN:
      status = self._InstantiateDomain(action)
    elif sourceElement.GetType()==ScolaModel.BLOCK:
      status = self._InstantiateBlock(action)
    elif sourceElement.GetType()==ScolaModel.PORT:
      status = self._InstantiatePort(action)
    elif sourceElement.GetType()==ScolaModel.ASSERTION:
      status = self._InstantiateAssertion(action)
    elif sourceElement.GetType()==ScolaModel.SCENARIO:
      status = self._InstantiateScenario(action)
    elif sourceElement.GetType()==ScolaModel.STATE:
      status = self._InstantiateState(action)
    elif sourceElement.GetType()==ScolaModel.TASK:
      status = self._InstantiateTask(action)
    elif sourceElement.GetType()==ScolaModel.TEST:
      status = self._InstantiateTestGateway(action)
    elif sourceElement.GetType()==ScolaModel.CASE:
      status = self._InstantiateGatewayCase(action)
    elif sourceElement.GetType()==ScolaModel.CHOICE:
      status = self._InstantiateChoiceGateway(action)
    elif sourceElement.GetType()==ScolaModel.BRANCH:
      status = self._InstantiateGatewayBranch(action)
    elif sourceElement.GetType()==ScolaModel.FORK:
      status = self._InstantiateForkGateway(action)
    elif sourceElement.GetType()==ScolaModel.JOIN:
      status = self._InstantiateJoinGateway(action)
    elif sourceElement.GetType()==ScolaModel.SPLIT:
      status = self._InstantiateSplitGateway(action)
    elif sourceElement.GetType()==ScolaModel.MERGE:
      status = self._InstantiateMergeGateway(action)
    elif sourceElement.GetType()==ScolaModel.MEET:
      status = self._InstantiateMeetGateway(action)
    elif sourceElement.GetType()==ScolaModel.NEXT:
      status = self._InstantiateNextDirective(action)
    elif sourceElement.GetType()==ScolaModel.FAIL:
      status = self._InstantiateFailInstruction(action)
    elif sourceElement.GetType()==ScolaModel.SET:
      status = self._InstantiateSetInstruction(action)
    elif sourceElement.GetType()==ScolaModel.NEW_PORT:
      status = self._InstantiateNewPortInstruction(action)
    elif sourceElement.GetType()==ScolaModel.NEW_BLOCK:
      status = self._InstantiateNewBlockInstruction(action)
    elif sourceElement.GetType()==ScolaModel.DELETE:
      status = self._InstantiateDeleteInstruction(action)
    elif sourceElement.GetType()==ScolaModel.MOVE:
      status = self._InstantiateMoveInstruction(action)
    elif sourceElement.GetType()==ScolaModel.SEND:
      status = self._InstantiateSendInstruction(action)
    elif sourceElement.GetType()==ScolaModel.RECEIVE:
      status = self._InstantiateReceiveInstruction(action)
    elif sourceElement.GetType()==ScolaModel.CLONE:
      status = self._InstantiateCloneInstruction(action)
    elif sourceElement.GetType()==ScolaModel.IF:
      status = self._InstantiateIfInstruction(action)
    elif sourceElement.GetType()==ScolaModel.BEGIN:
      status = self._InstantiateInstructionBlock(action)
    elif sourceElement.GetType()==ScolaModel.PRINT:
      status = self._InstantiatePrintInstruction(action)
    else:
      status = ScolaRewriterAction.FAILED
    return status

  def _InstantiateModel(self, action):
    sourceModel = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    self._PushInstantiateActionsForElements(sourceModel, targetContainer)
    return ScolaRewriterAction.COMPLETED

  def _PushInstantiateActionsForElements(self, sourceContainer, targetContainer):
    for sourceElement in reversed(sourceContainer.GetElements()):
      self._PushInstantiateActionForElement(sourceElement, targetContainer)

  def _PushInstantiateActionForElement(self, sourceElement, targetContainer):
    action = ScolaRewriterAction.NewInstantiateElement(sourceElement, targetContainer)
    self._PushAction(action)

  def _InstantiateInstance(self, action):
    sourceInstance = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    if sourceInstance.GetEventualType()==ScolaModel.BLOCK:
      targetBlock = ScolaBlock(sourceInstance.GetIdentifier(), targetContainer)
      sourceClass = sourceInstance.GetClassDeclaration()
      self._PushInstantiateActionsForElements(sourceInstance, targetBlock)
      self._InstantiateElementsOfInheritedClasses(sourceClass, targetBlock)
    elif sourceInstance.GetEventualType()==ScolaModel.SCENARIO:
      targetScenario = ScolaScenario(sourceInstance.GetIdentifier(), targetContainer)
      if sourceInstance.GetAsPath()!=None:
        targetScenario.SetAsPath(ScolaPath.Copy(sourceInstance.GetAsPath()))
      sourceClass = sourceInstance.GetClassDeclaration()
      self._PushInstantiateActionsForElements(sourceInstance, targetContainer)
      self._InstantiateElementsOfInheritedClasses(sourceClass, targetScenario)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateElementsOfInheritedClasses(self, sourceClass, targetContainer):
    self._PushInstantiateActionsForElements(sourceClass, targetContainer)
    inheritedClass = sourceClass.GetInheritedClass()
    if inheritedClass==None:
      return
    self._InstantiateElementsOfInheritedClasses(inheritedClass, targetContainer)

  def _InstantiateClonesDirective(self, action):
    sourceClonesDirective = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    pathToClonedElement = sourceClonesDirective.GetPathToClonedElement()
    targetContainerPath = targetContainer.GetAbsolutePath()
    actualPathToClonedElement = ScolaPath.Merge(targetContainerPath, pathToClonedElement)
    clonedElement = self._targetModel.AccessNamedElement(actualPathToClonedElement)
    if clonedElement==None:
      self.RaiseError("Attempt to clone an undeclared element")
      return ScolaRewriterAction.FAILED
    if clonedElement.GetType()==ScolaModel.BLOCK:
      targetBlock = ScolaBlock(sourceClonesDirective.GetIdentifier(), targetContainer)
      self._PushInstantiateActionsForElements(sourceClonesDirective, targetBlock)
      self._PushInstantiateActionsForElements(clonedElement, targetBlock)
    elif clonedElement.GetType()==ScolaModel.SCENARIO:
      targetScenario = ScolaScenario(sourceClonesDirective.GetIdentifier(), targetContainer)
      if sourceClonesDirective.GetAsPath()!=None:
        targetScenario.SetAsPath(ScolaPath.Copy(sourceClonesDirective.GetAsPath()))
      self._PushInstantiateActionsForElements(sourceClonesDirective, targetScenario)
      self._PushInstantiateActionsForElements(clonedElement, targetScenario)
    else:
      self.RaiseError("Attempt to clone a non cloneable element")
      return ScolaRewriterAction.FAILED
    return ScolaRewriterAction.COMPLETED

  def _InstantiateInheritsDirective(self, action):
    sourceInherits = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    pathToInheritedElement = sourceInherits.GetPathToInheritedElement()
    if pathToInheritedElement==None:
      return ScolaRewriterAction.COMPLETED
    targetContainerPath = targetContainer.GetAbsolutePath()
    actualPathToInheritedElement = ScolaPath.Merge(targetContainerPath, pathToInheritedElement)
    inheritedElement = self._targetModel.AccessNamedElement(actualPathToInheritedElement)
    if inheritedElement==None:
      self.RaiseError("Attempt to inherit an undeclared element")
      return ScolaRewriterAction.FAILED
    self._PushInstantiateActionsForElements(inheritedElement, targetContainer)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateDomain(self, action):
    sourceDomain = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetDomain = ScolaDomain(sourceDomain.GetIdentifier(), targetContainer)
    targetDomain.SetDomain(sourceDomain.GetDomain())
    return ScolaRewriterAction.COMPLETED

  def _InstantiateBlock(self, action):
    sourceBlock = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetBlock = ScolaBlock(sourceBlock.GetIdentifier(), targetContainer)
    self._PushInstantiateActionsForElements(sourceBlock, targetBlock)
    return ScolaRewriterAction.COMPLETED

  def _InstantiatePort(self, action):
    sourcePort = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourceBaseType = sourcePort.GetBaseType()
    if sourceBaseType.GetType()==ScolaBaseType.DOMAIN:
      identifier = sourceBaseType.GetIdentifier()
      path = ScolaPath.FromString(identifier)
      targetDomain = self._targetModel.AccessNamedElement(path)
      if targetDomain==None or targetDomain.GetType()!=ScolaModel.DOMAIN:
        self.RaiseError("Unable to retrieve domain %s" % identifier)
        return ScolaRewriterAction.FAILED
      targetBaseType = targetDomain.GetDomain()
    else:
      targetBaseType = sourceBaseType
    sourceInitialization = sourcePort.GetInitialization()
    targetInitialization = self._CopyExpression(sourceInitialization)
    targetPort = ScolaPort(targetBaseType, sourcePort.GetIdentifier(), targetInitialization, targetContainer)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateAssertion(self, action):
    sourceAssertion = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetAssertion = ScolaAssertion(sourceAssertion.GetIdentifier(), targetContainer)
    self._PushInstantiateActionsForElements(sourceAssertion, targetAssertion)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateScenario(self, action):
    sourceScenario = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetScenario = ScolaScenario(sourceScenario.GetIdentifier(), targetContainer)
    scenarioAsPath = sourceScenario.GetAsPath()
    if scenarioAsPath!=None:
      targetScenario.SetAsPath(ScolaPath.Copy(scenarioAsPath))
    self._PushInstantiateActionsForElements(sourceScenario, targetScenario)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateState(self, action):
    sourceState = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetState = ScolaState(sourceState.GetIdentifier(), targetContainer)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateTask(self, action):
    sourceTask = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetTask = ScolaTask(sourceTask.GetIdentifier(), targetContainer)
    self._PushInstantiateActionsForElements(sourceTask, targetTask)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateTestGateway(self, action):
    sourceGateway = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetGateway = ScolaTestGateway(sourceGateway.GetIdentifier(), targetContainer)
    self._PushInstantiateActionsForElements(sourceGateway, targetGateway)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateGatewayCase(self, action):
    sourceCase = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourceCondition = sourceCase.GetCondition()
    targetCondition = self._CopyExpression(sourceCondition)
    targetCase = ScolaGatewayCase(sourceCase.GetIdentifier(), targetCondition, targetContainer)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateChoiceGateway(self, action):
    sourceGateway = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetGateway = ScolaChoiceGateway(sourceGateway.GetIdentifier(), targetContainer)
    self._PushInstantiateActionsForElements(sourceGateway, targetGateway)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateGatewayBranch(self, action):
    sourceBranch = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetBranch = ScolaGatewayBranch(sourceBranch.GetIdentifier(), targetContainer)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateForkGateway(self, action):
    sourceGateway = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetGateway = ScolaForkGateway(sourceGateway.GetIdentifier(), targetContainer)
    self._PushInstantiateActionsForElements(sourceGateway, targetGateway)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateJoinGateway(self, action):
    sourceGateway = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetGateway = ScolaJoinGateway(sourceGateway.GetIdentifier(), targetContainer)
    self._PushInstantiateActionsForElements(sourceGateway, targetGateway)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateSplitGateway(self, action):
    sourceGateway = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetGateway = ScolaSplitGateway(sourceGateway.GetIdentifier(), targetContainer)
    self._PushInstantiateActionsForElements(sourceGateway, targetGateway)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateMergeGateway(self, action):
    sourceGateway = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetGateway = ScolaMergeGateway(sourceGateway.GetIdentifier(), targetContainer)
    self._PushInstantiateActionsForElements(sourceGateway, targetGateway)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateMeetGateway(self, action):
    sourceGateway = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetGateway = ScolaMeetGateway(sourceGateway.GetIdentifier(), targetContainer)
    self._PushInstantiateActionsForElements(sourceGateway, targetGateway)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateNextDirective(self, action):
    sourceDirective = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourcePath = sourceDirective.GetSourcePath()
    targetPath = sourceDirective.GetTargetPath()
    absoluteContainerPath = targetContainer.GetAbsolutePath()
    absoluteSourcePath = ScolaPath.Merge(absoluteContainerPath, sourcePath)
    sourceElement = self._targetModel.AccessNamedElement(absoluteSourcePath)
    absoluteTargetPath = ScolaPath.Merge(absoluteContainerPath, targetPath)
    targetElement = self._targetModel.AccessNamedElement(absoluteTargetPath)
    if sourceElement==None or targetElement==None:
      if self._noPostponing:
        if sourceElement==None:
          self.RaiseError("Invalid source element " + absoluteSourcePath.ToString())
        if targetElement==None:
          self.RaiseError("Invalid target element " + absoluteTargetPath.ToString())
        return ScolaRewriterAction.FAILED
      return ScolaRewriterAction.POSTPONED
    if not self._IsAllowedSourceElement(sourceElement):
      self.RaiseError("Invalid source element " + absoluteSourcePath.ToString())
      return ScolaRewriterAction.FAILED
    if not self._IsAllowedTargetElement(targetElement):
      self.RaiseError("Invalid target element " + absoluteTargetPath.ToString())
      return ScolaRewriterAction.FAILED
    sourceElement.SetNextElement(targetElement)
    if targetElement.GetType()==ScolaModel.STATE:
      targetElement.SetInitial(False)
    ScolaNextDirective(sourcePath, targetPath, targetContainer) # useful only if the compiled model is to be reused
    return ScolaRewriterAction.COMPLETED

  def _IsAllowedSourceElement(self, element):
    if element.GetType()==ScolaModel.STATE:
      return True
    if element.GetType()==ScolaModel.TASK:
      return True
    if element.GetType()==ScolaModel.JOIN:
      return True
    if element.GetType()==ScolaModel.MERGE:
      return True
    if element.GetType()==ScolaModel.CASE:
      return True
    if element.GetType()==ScolaModel.BRANCH:
      gateway = element.GetContainer()
      if gateway.GetType()==ScolaModel.CHOICE:
        return True
      if gateway.GetType()==ScolaModel.FORK:
        return True
      if gateway.GetType()==ScolaModel.SPLIT:
        return True
      if gateway.GetType()==ScolaModel.MEET:
        return True
    return False

  def _IsAllowedTargetElement(self, element):
    if element.GetType()==ScolaModel.STATE:
      return True
    if element.GetType()==ScolaModel.TASK:
      return True
    if element.GetType()==ScolaModel.TEST:
      return True
    if element.GetType()==ScolaModel.CHOICE:
      return True
    if element.GetType()==ScolaModel.FORK:
      return True
    if element.GetType()==ScolaModel.SPLIT:
      return True
    if element.GetType()==ScolaModel.BRANCH:
      gateway = element.GetContainer()
      if gateway.GetType()==ScolaModel.JOIN:
        return True
      if gateway.GetType()==ScolaModel.MERGE:
        return True
      if gateway.GetType()==ScolaModel.MEET:
        return True
    return False

  def _InstantiateFailInstruction(self, action):
    sourceInstruction = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetInstruction = ScolaFailInstruction(targetContainer)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateSetInstruction(self, action):
    sourceInstruction = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    if targetContainer.GetType()==ScolaModel.BLOCK:
      systemPath = targetContainer.GetAbsolutePath()
      self._evaluator.SetSystemPath(systemPath)
      path = self._evaluator.EvalExpressionAsPath(sourceInstruction.GetPathExpression())
      self._evaluator.SetSystemPath(None)
      targetPort = self._targetModel.AccessNamedElement(path)
      if targetPort==None or targetPort.GetType()!=ScolaModel.PORT:
        self.RaiseError("No port %s" % path.ToString())
        return ScolaRewriterAction.FAILED
      sourceValueExpression = sourceInstruction.GetValueExpression()
      targetValueExpression = self._CopyExpression(sourceValueExpression)
      targetPort.SetInitialization(targetValueExpression)
    else:
      sourcePathExpression = sourceInstruction.GetPathExpression()
      targetPathExpression = self._CopyExpression(sourcePathExpression)
      sourceValueExpression = sourceInstruction.GetValueExpression()
      targetValueExpression = self._CopyExpression(sourceValueExpression)
      targetInstruction = ScolaSetInstruction(targetPathExpression, targetValueExpression, targetContainer)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateNewPortInstruction(self, action):
    sourceInstruction = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetBaseType = sourceInstruction.GetBaseType()
    sourcePathExpression = sourceInstruction.GetPathExpression()
    targetPathExpression = self._CopyExpression(sourcePathExpression)
    sourceInitializationExpression = sourceInstruction.GetInitializationExpression()
    targetInitializationExpression = self._CopyExpression(sourceInitializationExpression)
    targetInstruction = ScolaNewPortInstruction(targetBaseType, targetPathExpression, targetInitializationExpression, targetContainer)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateNewBlockInstruction(self, action):
    sourceInstruction = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourceBlockClass = sourceInstruction.GetBlockClass()
    if sourceBlockClass==None:
      targetBlockClass = None
    else:
      identifier = sourceBlockClass.GetIdentifier()
      targetBlockClass = self._targetModel.LookForNamedElement(identifier)
      if targetBlockClass==None or targetBlockClass.GetType()!=ScolaModel.CLASS or targetBlockClass.GetEventualType()!=ScolaModel.BLOCK:
        self.RaiseError("Absent or invalid class " + identifier)
        return ScolaRewriterAction.FAILED
    sourcePathExpression = sourceInstruction.GetPathExpression()
    targetPathExpression = self._CopyExpression(sourcePathExpression)
    targetInstruction = ScolaNewBlockInstruction(targetBlockClass, targetPathExpression, targetContainer)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateDeleteInstruction(self, action):
    sourceInstruction = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourcePathExpression = sourceInstruction.GetPathExpression()
    targetPathExpression = self._CopyExpression(sourcePathExpression)
    targetInstruction = ScolaDeleteInstruction(targetPathExpression, targetContainer)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateMoveInstruction(self, action):
    sourceInstruction = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourceSourcePathExpression = sourceInstruction.GetPathExpression()
    targetSourcePathExpression = self._CopyExpression(sourceSourcePathExpression)
    sourceTargetPathExpression = sourceInstruction.GetTargetPathExpression()
    targetTargetPathExpression = self._CopyExpression(sourceTargetPathExpression)
    targetInstruction = ScolaMoveInstruction(targetSourcePathExpression, targetTargetPathExpression, targetContainer)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateSendInstruction(self, action):
    sourceInstruction = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourceSourcePathExpression = sourceInstruction.GetPathExpression()
    targetSourcePathExpression = self._CopyExpression(sourceSourcePathExpression)
    sourceReceiverPathExpression = sourceInstruction.GetReceiverPathExpression()
    targetReceiverPathExpression = self._CopyExpression(sourceReceiverPathExpression)
    targetInstruction = ScolaSendInstruction(targetSourcePathExpression, targetReceiverPathExpression, targetContainer)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateReceiveInstruction(self, action):
    sourceInstruction = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourceSourcePathExpression = sourceInstruction.GetPathExpression()
    targetSourcePathExpression = self._CopyExpression(sourceSourcePathExpression)
    sourceSenderPathExpression = sourceInstruction.GetSenderPathExpression()
    targetSenderPathExpression = self._CopyExpression(sourceSenderPathExpression)
    sourceTargetPathExpression = sourceInstruction.GetTargetPathExpression()
    targetTargetPathExpression = self._CopyExpression(sourceTargetPathExpression)
    targetInstruction = ScolaReceiveInstruction(targetSourcePathExpression, targetSenderPathExpression, targetTargetPathExpression, targetContainer)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateCloneInstruction(self, action):
    sourceInstruction = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourceSourcePathExpression = sourceInstruction.GetPathExpression()
    targetSourcePathExpression = self._CopyExpression(sourceSourcePathExpression)
    sourceTargetPathExpression = sourceInstruction.GetTargetPathExpression()
    targetTargetPathExpression = self._CopyExpression(sourceTargetPathExpression)
    targetInstruction = ScolaCloneInstruction(targetSourcePathExpression, targetTargetPathExpression, targetContainer)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateIfInstruction(self, action):
    sourceInstruction = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourceCondition = sourceInstruction.GetCondition()
    targetCondition = self._CopyExpression(sourceCondition)
    targetInstruction = ScolaIfInstruction(targetCondition, targetContainer)
    self._PushInstantiateActionsForElements(sourceInstruction, targetInstruction)
    return ScolaRewriterAction.COMPLETED

  def _InstantiateInstructionBlock(self, action):
    sourceInstruction = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetInstruction = ScolaInstructionBlock(targetContainer)
    self._PushInstantiateActionsForElements(sourceInstruction, targetInstruction)
    return ScolaRewriterAction.COMPLETED

  def _InstantiatePrintInstruction(self, action):
    sourceInstruction = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourceExpression = sourceInstruction.GetPathExpression()
    targetExpression = self._CopyExpression(sourceExpression)
    targetInstruction = ScolaPrintInstruction(targetExpression, targetContainer)
    return ScolaRewriterAction.COMPLETED

  def _CopyExpression(self, expression):
    if expression.GetType()==ScolaExpression.NONE:
      return ScolaExpression.NewNone()
    if expression.GetType()==ScolaExpression.REFERENCE:
      value = expression.GetValue()
      return ScolaExpression.NewReference(value.GetValueAsSymbol())
    if expression.GetType()==ScolaExpression.SYMBOL:
      value = expression.GetValue()
      symbol = value.GetValueAsSymbol()
      constant = ScolaUniverse.LookForSymbolicConstantOfDomains(symbol)
      if constant!=None:
        copy = ScolaExpression.NewSymbol(symbol)
        copy.SetValue(constant)
        return copy
      return ScolaExpression.NewReference(symbol)
    if expression.GetType()==ScolaExpression.TRUE:
      return ScolaExpression.NewTrue()
    if expression.GetType()==ScolaExpression.FALSE:
      return ScolaExpression.NewFalse()
    if expression.GetType()==ScolaExpression.INTEGER:
      value = expression.GetValue()
      return ScolaExpression.NewInteger(value.GetValueAsInteger())
    if expression.GetType()==ScolaExpression.REAL:
      value = expression.GetValue()
      return ScolaExpression.NewReal(value.GetValueAsReal())
    if expression.GetType()==ScolaExpression.STRING:
      value = expression.GetValue()
      return ScolaExpression.NewString(value.GetValueAsString())
    copy = ScolaExpression(expression.GetType())
    for child in expression.GetChildren():
      childCopy = self._CopyExpression(child)
      copy.AddChild(childCopy)
    return copy

  def _SetSystemPaths(self, element, path):
    if element.GetType()==ScolaModel.MODEL:
      self._SetSystemPathsOfElements(element, path)
    elif element.GetType()==ScolaModel.CLASS:
      pass
    elif element.GetType()==ScolaModel.INSTANCE:
      pass
    elif element.GetType()==ScolaModel.CLONES:
      pass
    elif element.GetType()==ScolaModel.INHERITS:
      pass
    elif element.GetType()==ScolaModel.DOMAIN:
      pass
    elif element.GetType()==ScolaModel.BLOCK:
      path.AppendIdentifier(element.GetIdentifier())
      self._SetSystemPathsOfElements(element, path)
      path.RemoveLastIdentifier()
    elif element.GetType()==ScolaModel.PORT:
      element.SetSystemPath(ScolaPath.Copy(path))
    elif element.GetType()==ScolaModel.ASSERTION:
      self._SetSystemPathsOfElements(element, path)
    elif element.GetType()==ScolaModel.SCENARIO:
      if element.GetAsPath()==None:
        self._SetSystemPathsOfElements(element, path)
      else:
        localPath = ScolaPath.Merge(path, element.GetAsPath())
        self._SetSystemPathsOfElements(element, localPath)
        asElement = self._targetModel.AccessNamedElement(localPath)
        if asElement==None or asElement.GetType()!=ScolaModel.BLOCK:
          self.RaiseError("No block %s to associate to scenario %s" % (localPath.ToString(), element.GetAbsolutePath().ToString()))
          return
    elif element.GetType()==ScolaModel.STATE:
      pass
    elif element.GetType()==ScolaModel.TASK:
      element.SetSystemPath(ScolaPath.Copy(path))
      self._SetSystemPathsOfElements(element, path)
    elif element.GetType()==ScolaModel.TEST:
      element.SetSystemPath(ScolaPath.Copy(path))
      self._SetSystemPathsOfElements(element, path)
    elif element.GetType()==ScolaModel.CASE:
      element.SetSystemPath(ScolaPath.Copy(path))
    elif element.GetType()==ScolaModel.CHOICE:
      element.SetSystemPath(ScolaPath.Copy(path))
      self._SetSystemPathsOfElements(element, path)
    elif element.GetType()==ScolaModel.BRANCH:
      element.SetSystemPath(ScolaPath.Copy(path))
    elif element.GetType()==ScolaModel.FORK:
      element.SetSystemPath(ScolaPath.Copy(path))
      self._SetSystemPathsOfElements(element, path)
    elif element.GetType()==ScolaModel.JOIN:
      element.SetSystemPath(ScolaPath.Copy(path))
      self._SetSystemPathsOfElements(element, path)
    elif element.GetType()==ScolaModel.SPLIT:
      element.SetSystemPath(ScolaPath.Copy(path))
      self._SetSystemPathsOfElements(element, path)
    elif element.GetType()==ScolaModel.MERGE:
      element.SetSystemPath(ScolaPath.Copy(path))
      self._SetSystemPathsOfElements(element, path)
    elif element.GetType()==ScolaModel.MEET:
      element.SetSystemPath(ScolaPath.Copy(path))
      self._SetSystemPathsOfElements(element, path)
    elif element.GetType()==ScolaModel.NEXT:
      pass
    elif element.GetType()==ScolaModel.SET:
      element.SetSystemPath(ScolaPath.Copy(path))
    elif element.GetType()==ScolaModel.NEW_PORT:
      element.SetSystemPath(ScolaPath.Copy(path))
    elif element.GetType()==ScolaModel.NEW_BLOCK:
      element.SetSystemPath(ScolaPath.Copy(path))
    elif element.GetType()==ScolaModel.DELETE:
      element.SetSystemPath(ScolaPath.Copy(path))
    elif element.GetType()==ScolaModel.MOVE:
      element.SetSystemPath(ScolaPath.Copy(path))
    elif element.GetType()==ScolaModel.SEND:
      element.SetSystemPath(ScolaPath.Copy(path))
    elif element.GetType()==ScolaModel.RECEIVE:
      element.SetSystemPath(ScolaPath.Copy(path))
    elif element.GetType()==ScolaModel.CLONE:
      element.SetSystemPath(ScolaPath.Copy(path))
    elif element.GetType()==ScolaModel.IF:
      element.SetSystemPath(ScolaPath.Copy(path))
      self._SetSystemPathsOfElements(element, path)
    elif element.GetType()==ScolaModel.BEGIN:
      self._SetSystemPathsOfElements(element, path)
    elif element.GetType()==ScolaModel.PRINT:
      element.SetSystemPath(ScolaPath.Copy(path))

  def _SetSystemPathsOfElements(self, container, path):
    for element in container.GetElements():
      self._SetSystemPaths(element, path)
